@@ -48,8 +48,8 @@
 /* the file where we store the recently used items */
 #define GTK_RECENTLY_USED_FILE	".recently-used.xbel"
 
-/* a poll every two seconds should be enough */
-#define POLL_DELTA	2000
+/* a poll approximately every five seconds */
+#define POLL_DELTA      5
 
 /* return all items by default */
 #define DEFAULT_LIMIT	-1
@@ -57,6 +57,13 @@
 /* keep in sync with xdgmime */
 #define GTK_RECENT_DEFAULT_MIME	"application/octet-stream"
 
+typedef struct
+{
+  GSourceFunc func;
+  gpointer data;
+  GDestroyNotify notify;
+} ThreadsDispatch;
+
 typedef struct
 {
   gchar *name;
@@ -96,7 +103,6 @@ struct _GtkRecentManagerPrivate
 {
   gchar *filename;
 
-  guint is_screen_singleton : 1;
   guint is_dirty : 1;
   guint write_in_progress : 1;
   guint read_in_progress : 1;
@@ -104,8 +110,6 @@ struct _GtkRecentManagerPrivate
   gint limit;
   gint size;
 
-  GdkScreen *screen;
-  
   GBookmarkFile *recent_items;
   
   time_t last_mtime;
@@ -121,6 +125,7 @@ enum
   PROP_SIZE
 };
 
+static void           gtk_recent_manager_dispose      (GObject               *object);
 static void           gtk_recent_manager_finalize     (GObject               *object);
 
 static void           gtk_recent_manager_set_property (GObject               *object,
@@ -150,6 +155,8 @@ static void           gtk_recent_info_free            (GtkRecentInfo         *re
 
 static guint signal_changed = 0;
 
+static GtkRecentManager *recent_manager_singleton = NULL;
+
 G_DEFINE_TYPE (GtkRecentManager, gtk_recent_manager, G_TYPE_OBJECT)
 
 static void
@@ -190,6 +197,32 @@ gtk_recent_manager_error_quark (void)
   return g_quark_from_static_string ("gtk-recent-manager-error-quark");
 }
 
+static gboolean
+threads_dispatch (gpointer data)
+{
+  ThreadsDispatch *dispatch = data;
+  gboolean res = FALSE;
+
+  GDK_THREADS_ENTER ();
+
+  if (!g_source_is_destroyed (g_main_current_source ()))
+    res = dispatch->func (dispatch->data);
+
+  GDK_THREADS_LEAVE ();
+
+  return res;
+}
+
+static void
+threads_free (gpointer data)
+{
+  ThreadsDispatch *dispatch = data;
+
+  if (dispatch->notify)
+    dispatch->notify (dispatch->data);
+
+  g_slice_free (ThreadsDispatch, dispatch);
+}
 
 static void
 gtk_recent_manager_class_init (GtkRecentManagerClass *klass)
@@ -200,6 +233,7 @@ gtk_recent_manager_class_init (GtkRecentManagerClass *klass)
   
   gobject_class->set_property = gtk_recent_manager_set_property;
   gobject_class->get_property = gtk_recent_manager_get_property;
+  gobject_class->dispose = gtk_recent_manager_dispose;
   gobject_class->finalize = gtk_recent_manager_finalize;
   
   /**
@@ -278,6 +312,7 @@ static void
 gtk_recent_manager_init (GtkRecentManager *manager)
 {
   GtkRecentManagerPrivate *priv;
+  ThreadsDispatch *dispatch;
   
   priv = g_type_instance_get_private ((GTypeInstance *) manager,
   				      GTK_TYPE_RECENT_MANAGER);
@@ -286,19 +321,23 @@ gtk_recent_manager_init (GtkRecentManager *manager)
   priv->limit = DEFAULT_LIMIT;
   priv->size = 0;
   
-  priv->is_screen_singleton = FALSE;
   priv->is_dirty = FALSE;
   priv->write_in_progress = FALSE;
   priv->read_in_progress = FALSE;
 
-  priv->screen = NULL;
-
   priv->filename = g_build_filename (g_get_home_dir (),
 				     GTK_RECENTLY_USED_FILE,
 				     NULL);
-  priv->poll_timeout = g_timeout_add (POLL_DELTA,
-		  		      gtk_recent_manager_poll_timeout,
-				      manager);
+  
+  dispatch = g_slice_new (ThreadsDispatch);
+  dispatch->func = gtk_recent_manager_poll_timeout;
+  dispatch->data = manager;
+  dispatch->notify = NULL;
+  priv->poll_timeout = g_timeout_add_seconds_full (G_PRIORITY_DEFAULT + 30,
+                                                   POLL_DELTA,
+                                                   threads_dispatch,
+                                                   dispatch,
+                                                   threads_free);
 
   build_recent_items_list (manager);
 }
@@ -350,23 +389,32 @@ gtk_recent_manager_get_property (GObject               *object,
     }
 } 
 
+static void
+gtk_recent_manager_dispose (GObject *object)
+{
+  GtkRecentManager *manager = GTK_RECENT_MANAGER (object);
+  GtkRecentManagerPrivate *priv = manager->priv;
+
+  if (priv->poll_timeout)
+    {
+      g_source_remove (priv->poll_timeout);
+      priv->poll_timeout = 0;
+    }
+
+  G_OBJECT_CLASS (gtk_recent_manager_parent_class)->dispose (object);
+}
+
 static void
 gtk_recent_manager_finalize (GObject *object)
 {
   GtkRecentManager *manager = GTK_RECENT_MANAGER (object);
   GtkRecentManagerPrivate *priv = manager->priv;
 
-  /* remove the poll timeout */
-  if (priv->poll_timeout)
-    g_source_remove (priv->poll_timeout);
-  
-  if (priv->filename)
-    g_free (priv->filename);
+  g_free (priv->filename);
   
   if (priv->recent_items)
     g_bookmark_file_free (priv->recent_items);
-  
-  /* chain up parent's finalize method */  
+
   G_OBJECT_CLASS (gtk_recent_manager_parent_class)->finalize (object);
 }
 
@@ -399,10 +447,7 @@ gtk_recent_manager_real_changed (GtkRecentManager *manager)
 	}
 
       write_error = NULL;
-      g_bookmark_file_to_file (priv->recent_items,
-		               priv->filename,
-			       &write_error);
-
+      g_bookmark_file_to_file (priv->recent_items, priv->filename, &write_error);
       if (write_error)
         {
           filename_warning ("Attempting to store changes into `%s', "
@@ -455,6 +500,7 @@ gtk_recent_manager_poll_timeout (gpointer data)
 {
   GtkRecentManager *manager = GTK_RECENT_MANAGER (data);
   GtkRecentManagerPrivate *priv = manager->priv;
+  time_t now;
   struct stat stat_buf;
   int stat_res;
 
@@ -462,6 +508,11 @@ gtk_recent_manager_poll_timeout (gpointer data)
   if (priv->write_in_progress || priv->read_in_progress)
     return TRUE;
 
+  /* do not stat the file if we're within 5 seconds from the last stat() */
+  now = time (NULL);
+  if (now < priv->last_mtime + 5);
+    return TRUE;
+
   stat_res = g_stat (priv->filename, &stat_buf);
   if (stat_res < 0)
     {
@@ -477,7 +528,7 @@ gtk_recent_manager_poll_timeout (gpointer data)
       return TRUE;
     }
 
-  /* the file didn't change from the last poll(), so we bail out */
+  /* the file didn't change from the last poll, so we bail out */
   if (stat_buf.st_mtime == priv->last_mtime)
     return TRUE;
 
@@ -492,6 +543,7 @@ gtk_recent_manager_set_filename (GtkRecentManager *manager,
 				 const gchar      *filename)
 {
   GtkRecentManagerPrivate *priv;
+  ThreadsDispatch *dispatch;
   
   g_assert (GTK_IS_RECENT_MANAGER (manager));
   priv = manager->priv;
@@ -508,9 +560,16 @@ gtk_recent_manager_set_filename (GtkRecentManager *manager,
     }
 
   priv->filename = g_strdup (filename);
-  priv->poll_timeout = g_timeout_add (POLL_DELTA,
-		  		      gtk_recent_manager_poll_timeout,
-				      manager);
+
+  dispatch = g_slice_new (ThreadsDispatch);
+  dispatch->func = gtk_recent_manager_poll_timeout;
+  dispatch->data = manager;
+  dispatch->notify = NULL;
+  priv->poll_timeout = g_timeout_add_seconds_full (G_PRIORITY_DEFAULT + 30,
+                                                   POLL_DELTA,
+                                                   threads_dispatch,
+                                                   dispatch,
+                                                   threads_free);
 
   /* mark us clean, so that we can re-read the list
    * of recently used resources
@@ -615,8 +674,7 @@ build_recent_items_list (GtkRecentManager *manager)
  * each time something inside the list changes.
  *
  * #GtkRecentManager objects are expensive: be sure to create them only when
- * needed. You should use the gtk_recent_manager_new_for_screen() or the
- * gtk_recent_manager_get_default() functions instead.
+ * needed. You should use gtk_recent_manager_get_default() instead.
  *
  * Return value: A newly created #GtkRecentManager object.
  *
@@ -631,20 +689,21 @@ gtk_recent_manager_new (void)
 /**
  * gtk_recent_manager_get_default:
  *
- * Gets the recent manager for the default screen. See
- * gtk_recent_manager_get_for_screen().
+ * Gets a unique instance of #GtkRecentManager, that you can share
+ * in your application without caring about memory management. The
+ * returned instance will be freed when you application terminates.
  *
- * Return value: A unique #GtkRecentManager associated with the
- *   default screen. This recent manager is associated with the
- *   screen and can be used as long as the screen is open.
- *   Do not ref or unref it.
+ * Return value: A unique #GtkRecentManager. Do not ref or unref it.
  *
  * Since: 2.10
  */
 GtkRecentManager *
 gtk_recent_manager_get_default (void)
 {
-  return gtk_recent_manager_get_for_screen (gdk_screen_get_default ());
+  if (G_UNLIKELY (!recent_manager_singleton))
+    recent_manager_singleton = gtk_recent_manager_new ();
+
+  return recent_manager_singleton;
 }
 
 /**
@@ -665,70 +724,16 @@ gtk_recent_manager_get_default (void)
  *   and can be used as long as the screen is open. Do not ref or
  *   unref it.
  *
+ * Deprecated: 2.12: This function has been deprecated and should
+ *   not be used in newly written code. Calling this function is
+ *   equivalent to calling gtk_recent_manager_get_default().
+ *
  * Since: 2.10
  */
 GtkRecentManager *
 gtk_recent_manager_get_for_screen (GdkScreen *screen)
 {
-  GtkRecentManager *manager;
-
-  g_return_val_if_fail (GDK_IS_SCREEN (screen), NULL);
-  g_return_val_if_fail (!screen->closed, NULL);
-
-  manager = g_object_get_data (G_OBJECT (screen), "gtk-recent-manager-default");
-  if (!manager)
-    {
-      GtkRecentManagerPrivate *priv;
-      
-      manager = gtk_recent_manager_new ();
-      gtk_recent_manager_set_screen (manager, screen);
-
-      priv = manager->priv;
-      priv->is_screen_singleton = TRUE;
-
-      g_object_set_data (G_OBJECT (screen), I_("gtk-recent-manager-default"), manager);
-    }
-
-  return manager;
-}
-
-static void
-display_closed (GdkDisplay       *display,
-		gboolean          is_error,
-		GtkRecentManager *manager)
-{
-  GtkRecentManagerPrivate *priv = manager->priv;
-  GdkScreen *screen = priv->screen;
-  gboolean was_screen_singleton = priv->is_screen_singleton;
-
-  if (was_screen_singleton)
-    {
-      g_object_set_data (G_OBJECT (screen), I_("gtk-recent-manager-default"), NULL);
-      priv->is_screen_singleton = FALSE;
-    }
-
-  gtk_recent_manager_set_screen (manager, NULL);
-
-  if (was_screen_singleton)
-    g_object_unref (manager);
-}
-
-static void
-unset_screen (GtkRecentManager *manager)
-{
-  GtkRecentManagerPrivate *priv = manager->priv;
-  GdkDisplay *display;
-
-  if (priv->screen)
-    {
-      display = gdk_screen_get_display (priv->screen);
-
-      g_signal_handlers_disconnect_by_func (display,
-					    (gpointer) display_closed,
-					    manager);
-
-      priv->screen = NULL;
-    }
+  return gtk_recent_manager_get_default ();
 }
 
 /**
@@ -741,30 +746,16 @@ unset_screen (GtkRecentManager *manager)
  * storage.
  * 
  * Since: 2.10
+ *
+ * Deprecated: 2.12: This function has been deprecated and should
+ *   not be used in newly written code. Calling this function has
+ *   no effect.
  */
 void
 gtk_recent_manager_set_screen (GtkRecentManager *manager,
 			       GdkScreen        *screen)
 {
-  GtkRecentManagerPrivate *priv;
-  GdkDisplay *display;
 
-  g_return_if_fail (GTK_IS_RECENT_MANAGER (manager));
-  g_return_if_fail (screen == NULL || GDK_IS_SCREEN (screen));
-
-  priv = manager->priv;
-
-  unset_screen (manager);
-
-  if (screen)
-    {
-      display = gdk_screen_get_display (screen);
-
-      priv->screen = screen;
-
-      g_signal_connect (display, "closed",
-		        G_CALLBACK (display_closed), manager);
-    }
 }
 
 /**
@@ -820,11 +811,11 @@ gtk_recent_manager_get_limit (GtkRecentManager *manager)
  * Adds a new resource, pointed by @uri, into the recently used
  * resources list.
  *
- * This function automatically retrieving some of the needed
+ * This function automatically retrieves some of the needed
  * metadata and setting other metadata to common default values; it
  * then feeds the data to gtk_recent_manager_add_full().
  *
- * See gtk_recent_manager_add_full() if you want to explicitely
+ * See gtk_recent_manager_add_full() if you want to explicitly
  * define the metadata for the resource pointed by @uri.
  *
  * Return value: %TRUE if the new item was successfully added
@@ -845,7 +836,7 @@ gtk_recent_manager_add_item (GtkRecentManager  *manager,
   recent_data.display_name = NULL;
   recent_data.description = NULL;
   recent_data.mime_type = NULL;
-  
+
 #ifdef G_OS_UNIX
   if (has_case_prefix (uri, "file:/"))
     {
@@ -856,9 +847,9 @@ gtk_recent_manager_add_item (GtkRecentManager  *manager,
       if (filename)
         {
           mime_type = xdg_mime_get_mime_type_for_file (filename, NULL);
-          if (mime_type && *mime_type)
+          if (mime_type)
             recent_data.mime_type = g_strdup (mime_type);
-
+      
           g_free (filename);
         }
 
@@ -1101,7 +1092,7 @@ gtk_recent_manager_has_item (GtkRecentManager *manager,
   return g_bookmark_file_has_item (priv->recent_items, uri);
 }
 
-static gboolean
+static void
 build_recent_info (GBookmarkFile  *bookmarks,
 		   GtkRecentInfo  *info)
 {
@@ -1161,8 +1152,6 @@ build_recent_info (GBookmarkFile  *bookmarks,
     }
   
   g_strfreev (apps);
-  
-  return TRUE; 
 }
 
 /**
@@ -1189,7 +1178,6 @@ gtk_recent_manager_lookup_item (GtkRecentManager  *manager,
 {
   GtkRecentManagerPrivate *priv;
   GtkRecentInfo *info = NULL;
-  gboolean res;
   
   g_return_val_if_fail (GTK_IS_RECENT_MANAGER (manager), NULL);
   g_return_val_if_fail (uri != NULL, NULL);
@@ -1224,14 +1212,8 @@ gtk_recent_manager_lookup_item (GtkRecentManager  *manager,
   /* fill the RecentInfo structure with the data retrieved by our
    * parser object from the storage file 
    */
-  res = build_recent_info (priv->recent_items, info);
-  if (!res)
-    {
-      gtk_recent_info_free (info);
-      
-      return NULL;
-    }
- 
+  build_recent_info (priv->recent_items, info);
+
   return info;
 }
 
@@ -1327,42 +1309,17 @@ gtk_recent_manager_get_items (GtkRecentManager *manager)
   for (i = 0; i < uris_len; i++)
     {
       GtkRecentInfo *info;
-      gboolean res;
+      
+      if (priv->limit != -1 && i == priv->limit)
+        break;
       
       info = gtk_recent_info_new (uris[i]);
-      res = build_recent_info (priv->recent_items, info);
-      if (!res)
-        {
-          g_warning ("Unable to create a RecentInfo object for "
-                     "item with URI `%s'",
-                     uris[i]);
-          gtk_recent_info_free (info);
-	  
-          continue;
-        }
+      build_recent_info (priv->recent_items, info);
       
       retval = g_list_prepend (retval, info);
     }
   
   g_strfreev (uris);
-    
-  /* clamp the list, if a limit is present */
-  if ((priv->limit != -1) &&
-      (g_list_length (retval) > priv->limit))
-    {
-      GList *clamp, *l;
-      
-      clamp = g_list_nth (retval, priv->limit - 1);
-      
-      if (!clamp)
-        return retval;
-      
-      l = clamp->next;
-      clamp->next = NULL;
-      
-      g_list_foreach (l, (GFunc) gtk_recent_info_free, NULL);
-      g_list_free (l);
-    }
   
   return retval;
 }
@@ -1439,7 +1396,7 @@ gtk_recent_info_get_type (void)
   static GType info_type = 0;
   
   if (!info_type)
-    info_type = g_boxed_type_register_static ("GtkRecentInfo",
+    info_type = g_boxed_type_register_static (I_("GtkRecentInfo"),
     					      (GBoxedCopyFunc) gtk_recent_info_ref,
     					      (GBoxedFreeFunc) gtk_recent_info_unref);
   return info_type;
@@ -1734,11 +1691,9 @@ recent_app_info_free (RecentAppInfo *app_info)
   if (!app_info)
     return;
   
-  if (app_info->name)
-    g_free (app_info->name);
+  g_free (app_info->name);
   
-  if (app_info->exec)
-    g_free (app_info->exec);
+  g_free (app_info->exec);
   
   g_free (app_info);
 }
@@ -2002,7 +1957,7 @@ gtk_recent_info_get_icon (GtkRecentInfo *info,
   if (info->mime_type)
     retval = get_icon_for_mime_type (info->mime_type, size);
 
-  /* this should never fail */
+  /* this function should never fail */  
   if (!retval)
     {
       if (info->mime_type &&
@@ -2011,7 +1966,7 @@ gtk_recent_info_get_icon (GtkRecentInfo *info,
       else
         retval = get_icon_fallback (GTK_STOCK_FILE, size);
     }
-
+  
   return retval;
 }
 
@@ -2196,7 +2151,7 @@ get_uri_shortname_for_display (const gchar *uri)
       rest = get_method_string (uri, &method);
       local_file = g_filename_display_basename (rest);
       
-      name = g_strdup_printf ("%s: %s", method, local_file);
+      name = g_strconcat (method, ": ", local_file, NULL);
       
       g_free (local_file);
       g_free (method);
@@ -2249,7 +2204,9 @@ gtk_recent_info_get_short_name (GtkRecentInfo *info)
  * gtk_recent_info_get_uri_display:
  * @info: a #GtkRecentInfo
  *
- * Gets a displayable version of the resource's URI.
+ * Gets a displayable version of the resource's URI.  If the resource
+ * is local, it returns a local path; if the resource is not local,
+ * it returns the UTF-8 encoded content of gtk_recent_info_get_uri().
  *
  * Return value: a UTF-8 string containing the resource's URI or %NULL
  *
@@ -2258,18 +2215,28 @@ gtk_recent_info_get_short_name (GtkRecentInfo *info)
 gchar *
 gtk_recent_info_get_uri_display (GtkRecentInfo *info)
 {
-  gchar *filename, *filename_utf8;
+  gchar *retval;
   
   g_return_val_if_fail (info != NULL, NULL);
-  
-  filename = g_filename_from_uri (info->uri, NULL, NULL);
-  if (!filename)
-    return NULL;
-      
-  filename_utf8 = g_filename_to_utf8 (filename, -1, NULL, NULL, NULL);
-  g_free (filename);
 
-  return filename_utf8;
+  retval = NULL;
+  if (gtk_recent_info_is_local (info))
+    {
+      gchar *filename;
+
+      filename = g_filename_from_uri (info->uri, NULL, NULL);
+      if (!filename)
+        return NULL;
+      
+      retval = g_filename_to_utf8 (filename, -1, NULL, NULL, NULL);
+      g_free (filename);
+    }
+  else
+    {
+      retval = make_valid_utf8 (info->uri);
+    }
+
+  return retval;
 }
 
 /**
@@ -2390,5 +2357,21 @@ gtk_recent_info_has_group (GtkRecentInfo *info,
   return FALSE;
 }
 
+/*
+ * _gtk_recent_manager_sync:
+ * 
+ * Private function for synchronising the recent manager singleton.
+ */
+void
+_gtk_recent_manager_sync (void)
+{
+  if (recent_manager_singleton)
+    {
+      /* force a dump of the contents of the recent manager singleton */
+      recent_manager_singleton->priv->is_dirty = TRUE;
+      gtk_recent_manager_real_changed (recent_manager_singleton);
+    }
+}
+
 #define __GTK_RECENT_MANAGER_C__
 #include "gtkaliasdef.c"
